# SPDX-License-Identifier: GPL-2.0+

import os, sys
import ctypes
from pprint import pprint

os.environ["_GFORT2PY_TEST_FLAG"] = "1"

import numpy as np
import gfort2py as gf

import pytest

SO = f"./tests/proc_ptrs.{gf.lib_ext()}"
MOD = "./tests/proc_ptrs.mod"

x = gf.fFort(SO, MOD)


class TestProcPtrsMethods:
    def assertEqual(self, x, y):
        assert x == y

    @pytest.mark.skip
    def test_proc_ptr_ffunc(self):
        x.sub_null_proc_ptr()
        with pytest.raises(AttributeError) as cm:
            y = x.p_func_func_run_ptr(1)

        x.p_func_func_run_ptr = x.func_func_run
        y = x.p_func_func_run_ptr(1)
        self.assertEqual(y.result, 10)
        y = x.p_func_func_run_ptr(2)
        self.assertEqual(y.result, 20)

        y = x.func_proc_ptr(5)
        y2 = x.p_func_func_run_ptr(5)
        self.assertEqual(y.result, y2.result)

    @pytest.mark.skip
    def test_proc_ptr_ffunc2(self):
        x.sub_null_proc_ptr()
        with pytest.raises(AttributeError) as cm:
            y = x.p_func_func_run_ptr2(1)  # Allready set

        x.p_func_func_run_ptr2 = x.func_func_run
        y = x.p_func_func_run_ptr2(10)
        self.assertEqual(y.result, 100)

    @pytest.mark.skip
    def test_proc_update(self):
        x.sub_null_proc_ptr()
        x.p_func_func_run_ptr = x.func_func_run
        y = x.p_func_func_run_ptr(1)
        self.assertEqual(y.result, 10)

        x.p_func_func_run_ptr = x.func_func_run2
        y = x.p_func_func_run_ptr(1)
        self.assertEqual(y.result, 2)

    @pytest.mark.skipif(gf.utils.is_big_endian(), reason="Skip on big endian systems")
    def test_proc_func_arg(self):
        y = x.func_func_arg_dp(5, x.func_real)
        self.assertEqual(y.result, 500)

        y = x.func_func_arg(x.func_func_run)
        self.assertEqual(y.result, 10)

        y = x.func_func_arg(func=x.func_func_run)
        self.assertEqual(y.result, 10)

    def test_proc_func_arg_compile(self):
        fstr = """
                integer function test(x)
                    integer :: x

                    test = 3*x
                end function test

                """

        f = gf.compile(fstr)

        y = x.func_func_arg(f.test)
        self.assertEqual(y.result, 3)

    @pytest.mark.skip
    def test_proc_proc_func_arg(self):
        x.sub_null_proc_ptr()
        x.p_func_func_run_ptr = x.func_func_run

        y = x.proc_proc_func_arg(x.p_func_func_run_ptr)
        self.assertEqual(y.result, 90)



import subprocess
import os
import sys

platforms = [
    [
        "armhf",
        "arm-linux-gnueabihf",
        "arm32v7/ubuntu:22.04",
        "linux/arm/v7",
    ],
    [
        "ppc64le",
        "powerpc64le-linux-gnu",
        "ppc64le/ubuntu:22.04",
        "linux/ppc64le",
    ],
    [
        "s390x",
        "s390x-linux-gnu",
        "s390x/ubuntu:22.04",
        "linux/s390x",
    ],
    [
        "riscv64",
        "riscv64-linux-gnu",
        "riscv64/ubuntu:22.04",
        "linux/riscv64",
    ],
]


for p in platforms:
    NAME = p[0]
    TOOLCHAIN_NAME = p[1]
    DOCKER_CONTAINER = p[2]
    PLATFORM_NAME = p[3]

    subprocess.run(
        f"""docker run --name {TOOLCHAIN_NAME} --platform {PLATFORM_NAME} --interactive -v $(pwd):/gfort2py {DOCKER_CONTAINER} /bin/bash -c "
          apt-get update &&
          apt-get install -y cmake git python3 python-is-python3 python3-dev python3-venv python3-pip python3-numpy automake libc6-dev linux-libc-dev gcc gfortran &&
          git config --global --add safe.directory /gfort2py &&
          python -m pip install build wheel pytest dataclasses_json cpyparsing platformdirs "
          """,
        shell=True,
    )

    subprocess.run(f"docker commit {TOOLCHAIN_NAME} {TOOLCHAIN_NAME}", shell=True)
    subprocess.run(
        f"docker save -o docker_{TOOLCHAIN_NAME}.tar {TOOLCHAIN_NAME}", shell=True
    )



# SPDX-License-Identifier: GPL-2.0+

import os, sys
import ctypes
from pprint import pprint

os.environ["_GFORT2PY_TEST_FLAG"] = "1"

import numpy as np
import gfort2py as gf

import pytest

SO = f"./tests/gh56.{gf.lib_ext()}"
MOD = "./tests/gh56.mod"

x = gf.fFort(SO, MOD)


class Testgh56Methods:
    def test_gh56(self, capfd):
        y1 = x.get_array(10).result

        y2 = x.get_array(5).result

        assert len(y1["d"]) == 10
        assert len(y2["d"]) == 5

        assert np.array_equal(
            y1["d"],
            np.array(
                [1.0, 2.0, 3.0, 4.0, 5.0, 6.0, 7.0, 8.0, 9.0, 10.0], dtype=np.float32
            ),
        )

        assert np.array_equal(
            y2["d"], np.array([1.0, 2.0, 3.0, 4.0, 5.0], dtype=np.float32)
        )



# SPDX-License-Identifier: GPL-2.0+

import ctypes
import itertools
from pprint import pprint
import platform
import os
import subprocess
import sys

from .fUnary import run_unary

_TEST_FLAG = os.environ.get("_GFORT2PY_TEST_FLAG") is not None


def copy_array(src, dst, length, size):
    ctypes.memmove(
        dst,
        src,
        length * size,
    )


def resolve_other_args(obj, other_args, module, lib, fProc):
    """
    We want to iterate over the components of obj
    and if they are symbol_refs look them up in other_args.

    This way we can resolve attributes that are runtime set,
    for instance dimension(n) arrays, where n is a dummy argument

    For now limit to array bounds

    fProc is needed in case we have to call a user function. We can't
    import it as utils.py is imported by fProc.py itself
    """

    if obj.is_array():
        for i in itertools.chain(obj.sym.array_spec.lower, obj.sym.array_spec.upper):
            i = _resolve_arg(i, other_args, module, lib, fProc)

    if obj.is_char():
        obj.sym.ts.charlen = _resolve_arg(
            obj.sym.ts.charlen, other_args, module, lib, fProc
        )

    return obj


def _resolve_arg(arg, other_args, module, lib, fProc):
    if not hasattr(arg, "exp_type"):
        return arg

    if arg.exp_type == "CONSTANT":
        return arg
    elif arg.exp_type == "VARIABLE":
        # Sometimes we try re-resolving already resolved arguments
        # so skip if value is not a symbol_ref
        if hasattr(arg._saved_value, "ref"):
            ref = arg._saved_value.ref
            for j in other_args:
                if ref == j.symbol_ref:
                    arg.value = j.value
    elif arg.exp_type == "OP":
        # Unary operator
        op = arg.unary_op
        arg1 = _resolve_arg(arg.unary_args[0], other_args, module, lib, fProc)
        arg2 = _resolve_arg(arg.unary_args[1], other_args, module, lib, fProc)
        arg.value = run_unary(op, arg1.value, arg2.value)
    elif arg.exp_type == "FUNCTION":
        # User supplied function
        func_ref = arg.value.ref
        # Lookup function
        func_sym = module[func_ref]
        func = fProc(lib, func_sym, module)
        # Lookup args
        # TODO: Only works for single argument functions for now
        arg = _resolve_arg(module[arg.args.value.ref], other_args, module, lib, fProc)
        for a in other_args:
            if a.symbol_ref == arg.head.id:
                func_arg = a.value
                break

        arg.value = func(func_arg).result

    return arg


def library_ext():
    """
    Determine shared library extension for a current os_platform
    """
    os_platform = platform.system()
    if os_platform == "Darwin":
        return "dylib"
    elif os_platform == "Windows" or "CYGWIN" in os_platform:
        return "dll"
    else:
        return "so"


def fc_path():
    """
    Guess location of gfortran compiler
    """
    if "FC" in os.environ:
        return os.environ["FC"]

    os_platform = platform.system()
    if os_platform == "Darwin":
        # Homebrew location
        if os.path.exists("/usr/local/bin/gfortran"):
            return "/usr/local/bin/gfortran"

    cmd = "where" if os_platform == "Windows" else "which"
    x = os.path.normpath(
        subprocess.run([cmd, "gfortran"], capture_output=True).stdout.decode().strip()
    )

    x = x.split()

    fc = None
    if _TEST_FLAG and os_platform == "Windows":
        for i in x:
            if "Chocolatey" in i:
                # Hardcode the choice for testing
                fc = i.strip()
                break
    else:
        fc = x[0].strip()

    if fc is None:
        raise ValueError("Did not find a gfortran compilier")

    # Windows may return several possible paths
    return fc


def is_64bit():
    return platform.architecture()[0] == "64bit"


def is_big_endian():
    return sys.byteorder == "big"


def is_ppc64le():
    return platform.machine() == "ppc64le"



#!/usr/bin/env python
#  SPDX-License-Identifier: GPL-2.0+
import sys


def make_test_case(name):

    with open(f"{name}_test.py", "w") as f:
        str = f"""
# SPDX-License-Identifier: GPL-2.0+

import os, sys
import ctypes
from pprint import pprint

os.environ["_GFORT2PY_TEST_FLAG"] = "1"

import numpy as np
import gfort2py as gf

import pytest

SO = f"./tests/{name}.{{gf.lib_ext()}}"
MOD = "./tests/{name}.mod"

x = gf.fFort(SO,MOD)


class Test{name}Methods:
    pass

        """

        print(str, file=f)

    with open(f"{name}.f90", "w") as f:
        str = f"""   

! SPDX-License-Identifier: GPL-2.0+

module {name}

	use iso_fortran_env, only: output_unit, real128
	
	implicit none
	
	! Parameters
	integer, parameter :: dp = selected_real_kind(p=15)
	integer, parameter :: qp = selected_real_kind(p=30)
	integer, parameter :: lp = selected_int_kind(16)
	
	
	contains



end module {name}

        """
        print(str, file=f)


if __name__ == "__main__":
    for i in sys.argv[1:]:
        make_test_case(i)



import os, sys
import ctypes
from pprint import pprint
import platform

os.environ["_GFORT2PY_TEST_FLAG"] = "1"

import numpy as np
import gfort2py as gf

import pytest


class TestCompileMethods:
    def test_compile_nomod_str(self):
        fstr = """
            integer function myfunc(x,y)
                integer :: x,y
                myfunc = x+y
            end function myfunc
            """

        x = gf.compile(string=fstr)

        result = x.myfunc(1, 2)

        assert result.result == 3

    def test_compile_mod_str(self):
        fstr = """
            module abc
            contains
            integer function myfunc(x,y)
                integer :: x,y
                myfunc = x+y
            end function myfunc
            end module abc
            """

        x = gf.compile(string=fstr)

        result = x.myfunc(1, 2)

        assert result.result == 3

    def test_compile_nomod_file(self):
        x = gf.compile(file="tests/compile_nomod_test.f90")

        result = x.myfunc(1, 2)

        assert result.result == 3

    def test_compile_mod_file(self):
        x = gf.compile(file="tests/compile_mod_test.f90")

        result = x.myfunc(1, 2)

        assert result.result == 3



from utils import *


modname = "vars_modules"
libname = "t_vars_modules"
filename_f90 = "../t_vars_modules.f90"
filename_py = "../t_vars_modules_test.py"


def make_func(func_name, name, v, n):
    f = []
    f.append(fort_func_start.substitute(name=func_name, args="", result="x"))
    f.append(INDENT + fort_bool_var.substitute(var="x"))
    f.append(INDENT + fort_var_set.substitute(name="x", value=".false."))
    f.append(INDENT + fort_var_comp.substitute(name=name, value=f"{v}_{n}"))
    f.append(INDENT + fort_var_set.substitute(name="x", value=".true."))
    f.append(INDENT + fort_func_end.substitute(name=func_name))

    return "\n".join(f)


def create_ints():
    fort_strs = []
    fort_funcs = []
    py_strs = []
    values = [-1, 0, 1]

    for n, k in fort_int_kinds.items():
        for indv, v in enumerate(values):
            name = f"int_{n}_{indv}"

            fort_strs.append(
                fort_set_var.substitute(
                    type="integer", kind=n, var=name, value=f"{v}_{n}"
                )
            )

            py_strs.append(py_value_comp.substitute(name=name, value=v))

            v = 2 * v
            py_strs.append(py_value_set.substitute(name=name, value=v))
            py_strs.append(py_value_comp.substitute(name=name, value=v))

            func_name = f"check_{name}"
            py_strs.append(py_proc_call.substitute(proc=func_name, args=""))
            py_strs.append(py_proc_return.substitute())

            fort_funcs.append(make_func(func_name, name, v, n))

    return fort_strs, fort_funcs, py_strs


def create_reals():
    fort_strs = []
    fort_funcs = []
    py_strs = []
    values = [-3.140000104904175, 0, 3.140000104904175]

    for n, k in fort_real_kinds.items():
        for indv, v in enumerate(values):
            name = f"real_{n}_{indv}"

            fort_strs.append(
                fort_set_var.substitute(type="real", kind=n, var=name, value=f"{v}_{n}")
            )

            py_strs.append(py_value_comp.substitute(name=name, value=v))

            v = 2 * v
            py_strs.append(py_value_set.substitute(name=name, value=v))
            py_strs.append(py_value_comp.substitute(name=name, value=v))

            func_name = f"check_{name}"
            py_strs.append(py_proc_call.substitute(proc=func_name, args=""))
            py_strs.append(py_proc_return.substitute())

            fort_funcs.append(make_func(func_name, name, v, n))

    return fort_strs, fort_funcs, py_strs


with open(filename_f90, "w") as file_f90, open(filename_py, "w") as file_py:
    write_line(file_f90, fort_module_start.substitute(name=modname), 0)
    write_line(file_py, py_module_start.substitute(modname=modname, libname=libname), 0)

    f_ints, ffuncs_ints, py_ints = create_ints()
    write_lines(file_f90, f_ints, 1)
    write_line(file_py, py_proc.substitute(function="check_ints"), 1)
    write_lines(file_py, py_ints, 2)

    f_reals, ffuncs_reals, py_reals = create_reals()
    write_lines(file_f90, f_reals, 1)
    write_line(file_py, py_proc.substitute(function="check_reals"), 1)
    write_lines(file_py, py_reals, 2)

    write_line(file_f90, fort_module_mid.substitute(), 1)

    write_lines(file_f90, ffuncs_ints, 1)
    write_lines(file_f90, ffuncs_reals, 1)

    # print(fort_module_mid.substitute(),file=file_f90)
    write_line(file_f90, fort_module_end.substitute(name=modname), 0)



# SPDX-License-Identifier: GPL-2.0+
import ctypes
import numpy as np
import os
import sys
import platform

from .module_parse import module

from .fVar import fVar
from .fProc import fProc
from .fParameters import fParam
from .fCompile import compile_and_load, common_compile
from .utils import library_ext

_TEST_FLAG = os.environ.get("_GFORT2PY_TEST_FLAG") is not None


class fFort:
    _initialized = False

    def __init__(self, libname, mod_file, cache_folder=None):
        """
        Loads a gfortran module given by mod_file and saved in a
        shared library libname.

        cache_folder: If not None, sets the folder location to saved
        the cached module data to. If None uses appdirs ``user_cache_dir``
        location.

        Set to False to disable caching.

        """

        self._libname = libname
        self._load_lib()
        self._mod_file = mod_file
        self._module = module(self._mod_file, cache_folder=cache_folder)

        self._saved = {}
        self._initialized = True

    def _load_lib(self):
        """
        Handle differences between windows and linux
        """
        kwargs = {}

        if platform.system() == "Windows":
            self._libname = os.path.realpath(self._libname)
            if sys.version_info.major >= 3:
                if sys.version_info.minor >= 8:
                    os.add_dll_directory(os.path.dirname(self._libname))
                    kwargs["winmode"] = 0

        if not os.path.exists(self._libname):
            raise FileNotFoundError(f"Can't find {self._libname}")

        self._lib = ctypes.CDLL(self._libname, **kwargs)

    def keys(self):
        return self._module.keys()

    def __contains__(self, key):
        return key in self._module.keys()

    def __dir__(self):
        return self.keys()

    def __getattr__(self, key):
        k = key
        key = key.lower()

        if key in self.__dict__:
            return self.__dict__[key]

        if "_initialized" in self.__dict__:
            if self._initialized:
                if key not in self.keys():
                    raise AttributeError(f"{self._mod_file}  has no attribute {k}")

            if self._module[key].is_variable():
                if key not in self._saved:
                    self._saved[key] = fVar(self._module[key], allobjs=self._module)
                self._saved[key].in_dll(self._lib)
                return self._saved[key].value
            elif self._module[key].is_proc_pointer():
                # Must come before fProc
                if key not in self._saved:
                    self._saved[key] = fVar(self._module[key], allobjs=self._module)
                return self._saved[key]
            elif self._module[key].is_procedure():
                if key not in self._saved:
                    self._saved[key] = fProc(self._lib, self._module[key], self._module)
                return self._saved[key]
            elif self._module[key].is_parameter():
                return fParam(self._module[key]).value
            else:
                raise NotImplementedError(
                    f"Object type {self._module[key].flavor()} not implemented yet"
                )

    def __setattr__(self, key, value):
        k = key
        key = key.lower()

        if key in self.__dict__:
            self.__dict__[key] = value
            return

        if "_initialized" in self.__dict__:
            if self._initialized:
                if self._module[key].is_variable():
                    if key not in self._saved:
                        self._saved[key] = fVar(self._module[key], allobjs=self._module)
                    self._saved[key].in_dll(self._lib)
                    self._saved[key].value = value
                    return
                elif self._module[key].is_parameter():
                    raise AttributeError("Can not alter a parameter")
                elif self._module[key].is_proc_pointer():
                    if key not in self._saved:
                        self._saved[key] = fVar(self._module[key], allobjs=self._module)
                    self._saved[key].value = value
                    return
                else:
                    raise NotImplementedError(
                        f"Object type {self._module[key].flavor()} not implemented yet"
                    )

        self.__dict__[key] = value

    @property
    def __doc__(self):
        return f"MODULE={self._module.filename}"

    def __str__(self):
        return f"{self._module.filename}"


def mod_info(mod_file, *, load_only=False, json=False):
    """
    Returns a parsed data structure that describes the module

    pprint is recommened to help understand the nested structure.
    """
    return module(mod_file, load_only=load_only, json=json)


def lib_ext():
    """
    Determine shared library extension for the current OS
    """
    return library_ext()


def compile(
    string=None,
    *,
    file=None,
    FC=None,
    FFLAGS="-O2",
    LDLIBS="",
    LDFLAGS="",
    output=None,
    cache_folder=None,
):
    """
    Compiles and loads a snippet of Fortran code.

    Either provide the code as a string in the ``string``
    argument of a filename in the ``file`` argument.

    This code will then be converted into a Fortran module and
    compiled.

    FC specifies the Fortran compiler to be used. This
    must be some version of gfortran

    FFLAGS specifies Fortran compile options. This defaults
    to -O2. We will additionally insert flags for building
    shared libraries on the current platform.

    LDLIBS specifies any libraries to link against

    LDFLAGS specifies extra argument to pass to the linker
    (usually this is specifying the directory of where libraries are
    stored and passed with the -L option)

    output Path to store intermediate files. Defaults to None
    where files are stored in a temp folder. Otherwise
    stored in ``output`` folder.

    cahce_folder same as for fFort, specifies location to save cached
    mod data to.

    """

    library, mod_file = compile_and_load(
        string=string,
        file=file,
        FC=FC,
        FFLAGS=FFLAGS,
        LDLIBS=LDLIBS,
        LDFLAGS=LDFLAGS,
        output=output,
    )

    return fFort(library, mod_file, cache_folder=cache_folder)


# Does not work needs https://gcc.gnu.org/bugzilla/show_bug.cgi?id=47030 applied

# def common(
#     string=None,
#     gfort=None,
#     *,
#     FC=None,
#     FFLAGS="-O2",
#     LDLIBS="",
#     LDFLAGS="",
#     output=None,
#     cache_folder=None,
# ):
#     """
#     Compiles and loads a snippet of Fortran code for accessing common blocks.

#     Code must be provided as a valid bit of Fortran that declares the variables
#     and declares the common block, eg:

#     fstr = "
# 		integer ::  a_int,b_int,c_int
# 		common  /common_name/ a_int,b_int,c_int
#          "

#     This code will then be converted into a Fortran module and
#     compiled.

#     gfort specifies an already loaded instance of fFort contains the common block

#     FC specifies the Fortran compiler to be used. This
#     must be some version of gfortran

#     FFLAGS specifies Fortran compile options. This defaults
#     to -O2. We will additionally insert flags for building
#     shared libraries on the current platform.

#     LDLIBS specifies any libraries to link against

#     LDFLAGS specifies extra argument to pass to the linker
#     (usually this is specifying the directory of where libraries are
#     stored and passed with the -L option)

#     output Path to store intermediate files. Defaults to None
#     where files are stored in a temp folder. Otherwise
#     stored in ``output`` folder.

#     cache_folder same as for fFort, specifies location to save cached
#     mod data to.

#     """

#     library, mod_file = common_compile(
#         string=string,
#         gfort=gfort,
#         FC=FC,
#         FFLAGS=FFLAGS,
#         LDLIBS=LDLIBS,
#         LDFLAGS=LDFLAGS,
#         output=output,
#     )

#     return fFort(library, mod_file, cache_folder=cache_folder)



# SPDX-License-Identifier: GPL-2.0+

import os, sys
import ctypes
from pprint import pprint

os.environ["_GFORT2PY_TEST_FLAG"] = "1"

import numpy as np
import gfort2py as gf

import pytest

SO = f"./tests/gh55.{gf.lib_ext()}"
MOD = "./tests/gh55.mod"

x = gf.fFort(SO, MOD)


@pytest.mark.skip("Currently broken needs array descriptor for strings")
class Testgh55Methods:
    def test_func_str_return_array(self):

        size = 3
        res = x.return_char(size)

        return_str = res.result

        assert np.size(np.shape(return_str)) == ((2**size) - 1)
        assert return_str[-1] == "abcdefghij"



# SPDX-License-Identifier: GPL-2.0+

import os, sys
import ctypes
from pprint import pprint

os.environ["_GFORT2PY_TEST_FLAG"] = "1"

import numpy as np
import gfort2py as gf

import pytest

SO = f"./tests/isoc.{gf.lib_ext()}"
MOD = "./tests/isoc.mod"

x = gf.fFort(SO, MOD)


class TestISOC:
    def test_c_name(self):
        # This should fail as we access via the
        # Fortran not c name
        with pytest.raises(KeyError) as e:
            x.a_int_binc_c = 1

    def test_f_name(self):
        x.a_int = 5
        assert x.a_int == 5

    def test_function(self):
        res = x.func_bind_c(1)

        assert res.result == 2



import subprocess
import tempfile
import os
import platform
import random
import string
import shutil
import hashlib
import platformdirs
from pathlib import Path
import logging

from .utils import library_ext, fc_path

_TEST_FLAG = os.environ.get("_GFORT2PY_TEST_FLAG") is not None


def compile_and_load(
    string=None,
    file=None,
    FC=None,
    FFLAGS="-O2",
    LDLIBS="",
    LDFLAGS="",
    output=None,
):
    if string is None and file is None:
        raise AttributeError("Must set either string or file")

    if FC is None:
        FC = fc_path()

    if _TEST_FLAG is True:
        logging.debug(f"Found FC={FC}")
        r = subprocess.run([FC, "-v"], capture_output=True)
        logging.debug(r.stdout)
        logging.debug(r.stderr)
        logging.debug(f"Environ = {os.environ.get('FC')}")

    output_dir = output_folder(output)
    output_file = output_filename(file, output_dir)

    if string is not None:
        str2file(string, output_file)
    else:
        shutil.copy(file, output_file)
        moduleize_file(file, output_file)

    mname = mod_name(output_file)

    lib_name = library_name(mname)

    library(lib_name, output_file, output_dir, FC, FFLAGS, LDLIBS, LDFLAGS)

    return os.path.join(output_dir, lib_name), module_filename(mname, output_dir)


def common_compile(
    string=None,
    gfort=None,
    FC=None,
    FFLAGS="-O2",
    LDLIBS="",
    LDFLAGS="",
    output=None,
):
    if "\n" in string:
        string = string.split("\n")

    name = "c" + hashlib.md5(b"".join([i.encode() for i in string])).hexdigest()

    string = "\n".join([f"module {name}", *string, "contains", "end module"])

    lib_path = Path(os.path.realpath(gfort._lib._name))

    LDLIBS = (
        LDLIBS + f"-l:{lib_path.name}"
    )  # colon is needed to search for exact name not lib{name}
    LDFLAGS = LDFLAGS + f"-L{lib_path.parent}"

    FFLAGS += f" -Wl,-rpath='.',-rpath='{lib_path.parent}' -Wl,--no-undefined"

    return compile_and_load(
        string=string,
        FC=FC,
        FFLAGS=FFLAGS,
        LDLIBS=LDLIBS,
        LDFLAGS=LDFLAGS,
        output=output,
    )


class CompileError(Exception):
    pass


def shared_lib_flags():
    os_platform = platform.system()
    if os_platform == "Darwin":
        return ["-dynamiclib"]
    elif os_platform == "Windows":
        return ["-shared"]
    else:
        return ["-fPIC", "-shared"]


def library(lib, file, output, FC, FFLAGS, LDLIBS, LDFLAGS):
    local_file = os.path.basename(file)

    line = " ".join(
        [FC, FFLAGS, *shared_lib_flags(), LDFLAGS, LDLIBS, "-o", lib, local_file]
    )

    res = subprocess.check_output(
        line,
        stderr=subprocess.STDOUT,
        cwd=output,
        shell=True,
    )


def moduleize_file(file, output):
    string = []
    with open(file, "r") as f:
        for line in f.readlines():
            string.append(line.strip("\n"))

    str2file(string, output)


def str2file(string, output):
    string = moduleize(string)
    with open(output, "w") as f:
        f.write("\n".join(string))


def moduleize(string):
    if "\n" in string:
        string = string.split("\n")

    # Check if already a module
    count = 0
    for line in string:
        line = line.strip()
        if line.startswith("module"):
            count += 1
        if line.startswith("contains"):
            count += 1
        if line.startswith("end module"):
            count += 1

    if count == 3:
        return string  # Already a module

    if count != 0:
        raise ValueError("Malformed module")

    # Module names must start with a letter
    # hashing the contents means the name is static so
    # module_parse caching should speed things up
    name = "a" + hashlib.md5(b"".join([i.encode() for i in string])).hexdigest()

    return [f"module {name}\n", "contains\n", *string, "end module\n"]


def mod_name(file):
    with open(file, "r") as f:
        for line in f.readlines():
            line = line.lower()
            if line.strip().startswith("module "):
                n = line.split()
                name = n[n.index("module") + 1]
                if "!" in name:
                    return name[: name.index("!")]
                else:
                    return name

    raise ValueError(f"Could not determine module name for {file}")


def output_folder(output):
    if output is None:
        # return platformdirs.user_cache_dir("gfort2py")
        return tempfile.mkdtemp(prefix="gfort2py")
    else:
        return os.path.realpath(output)


def output_filename(file, folder):
    if file is None:
        f, name = tempfile.mkstemp(suffix=".f90", prefix="mod_", dir=folder)
        os.close(f)
        return name
    else:
        p = Path(file)
        return os.path.join(folder, f"mod_{p.name}")


def library_name(file):
    return f"lib{file}.{library_ext()}"


def module_filename(module_name, output_folder):
    return os.path.join(output_folder, f"{module_name}.mod")


def random_string(N):
    return "".join(random.choices(string.ascii_lowercase, k=N))
